from collections import Iterable
from pytz import timezone, utc

import datetime
import email.utils
import inspect
import re
import time


def native_str(x, encoding="utf-8", errors="ignore", default=None):
    if x:
        return x if isinstance(x, str) else x.decode(encoding, errors)
    else:
        return default


def now(tz_str="utc"):
    return datetime.datetime.now(timezone(tz_str))


def convert_from_utc(dt, tz_str):
    return dt.replace(tzinfo=utc).astimezone(timezone(tz_str))


def rfc_2822_timestamp(dt=None):
    if not dt:
        dt = datetime.datetime.now()
    return email.utils.formatdate(time.mktime(dt.timetuple()))


def convert_from_rfc_2822(timestamp):
    return datetime.datetime.fromtimestamp(email.utils.mktime_tz(email.utils.parsedate_tz(timestamp)), utc)


def int_parse(x, default=None):
    if not x:
        return x
    try:
        return int(x)
    except ValueError:
        return default


def trunc(s, length=-1, strip=True, ellipsis=False, convert_to_none=True):
    """
    Truncates a string to the given length. If strip is
    true then also strips the string.  If ellipsis is true then
    ellipsis are added to the end of the string.
    """
    if not s:
        if convert_to_none and s is not None:
            return None
        return s
    if strip:
        s = s.strip()
        if convert_to_none and not s:
            return None
    if length > 0:
        if ellipsis:
            length -= 3
            return (s[:length] + "...") if len(s) > length else s
        else:
            return s[:length]
    else:
        return s


def xstr(s):
    """
    Returns the given object as a string or if None then
    returns the empty string
    """
    return str(s) if s else ""


def combine_bytes(*args, separator=None):
    count = len(args)
    b = bytearray()
    for indx, val in enumerate(args):
        b += val
        if separator and indx + 1 < count:
            b += separator
    return b


def unique(l):
    """Returns a new list with all the unique elements in l"""
    return list(set(l))


def intersect(a, b):
    """Returns a new list of the intersection of a and b"""
    return set(a) & set(b)


def union(a, b):
    """Returns a new list of the union between a and b"""
    return set(a) | set(b)


def diff(a, b):
    """Returns a new list of the differences between a and b"""
    return set(b).difference(set(a))


def create_frozenset(obj):
    if not obj:
        return frozenset()
    if not isinstance(obj, str) and isinstance(obj, Iterable):
        return frozenset(obj)
    else:
        return frozenset([obj])


def create_list(obj):
    if not obj:
        return list()
    if not isinstance(obj, str) and isinstance(obj, Iterable):
        return list(obj)
    else:
        return [obj]


def create_set(obj):
    if not obj:
        return set()
    if not isinstance(obj, str) and isinstance(obj, Iterable):
        return set(obj)
    else:
        return {obj}


def hasmethod(obj, name):
    if hasattr(obj, name):
        method = getattr(obj, name)
        return inspect.ismethod(method)
    return False


def hasfunc(obj, name):
    if hasattr(obj, name):
        func = getattr(obj, name)
        return inspect.isfunction(func)
    return False


_REPLACE_UNDERSCORE = re.compile("_([a-z])")


def camel_match(match):
    return match.group(0)[1:].upper()


def camel_encode(s):
    return re.sub(_REPLACE_UNDERSCORE, camel_match, s)


_FIRST_CAP = re.compile("(.)([A-Z][a-z]+)")
_ALL_CAP = re.compile("([a-z0-9])([A-Z])")


def camel_decode(s):
    return _ALL_CAP.sub(r"\1_\2", _FIRST_CAP.sub(r"\1_\2", s)).lower()


import logging

from dorthy.settings import config

from .base import Session, InMemorySessionStore

logger = logging.getLogger(__name__)

session_store = None
if "web.session_store" in config:
    if config.web.session_store == "redis":
        from .redis import RedisSessionStore
        session_store = RedisSessionStore()
        logger.info("Using Session Store: redis")
    elif config.web.session_store == "db":
        from .db import DBSessionStore
        session_store = DBSessionStore()
        logger.info("Using Session Store: db")

if session_store is None:
    session_store = InMemorySessionStore()
    logger.info("Using Session Store: in-memory")



import logging
import os
import smtplib

from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from jinja2 import Environment, FileSystemLoader, TemplateNotFound

from dorthy.settings import config

logger = logging.getLogger(__name__)

default_from = config.mail.from_address
default_reply_to = config.mail.get("reply_to_address")
log_message = config.mail.enabled("log_message")

if config.mail.templates[0] == "/":
    _template_path = config.mail.templates
else:
    _template_path = (os.path.join(os.getcwd(), config.mail.templates))

template_env = Environment(loader=FileSystemLoader(_template_path))


def _get_template(template_name, extension):
    try:
        return template_env.get_template(template_name + "." + extension)
    except TemplateNotFound:
        return None


def send_message(to_address, from_address=default_from, subject=None,
                 reply_to=None, use_reply_to=True, text_msg=None, html_msg=None):

    assert text_msg or html_msg, "A message must be sent - html or text or both."

    msg = MIMEMultipart("alternative")
    if subject:
        msg["Subject"] = subject
    msg["From"] = from_address
    msg["To"] = to_address

    if text_msg:
        msg.attach(MIMEText(text_msg, "plain"))

    if html_msg:
        msg.attach(MIMEText(html_msg, "html"))

    if log_message:
        logger.info(msg.as_string())
    else:
        s = None
        try:
            s = smtplib.SMTP(host=config.mail.host, port=config.mail.get("port", 25))
            if config.mail.enabled("debug"):
                s.set_debuglevel(1)
            if config.mail.enabled("use_starttls"):
                s.starttls()

            s.login(config.mail.username, config.mail.password)

            if use_reply_to:
                if not reply_to:
                    reply_to = default_reply_to if default_reply_to else from_address
                msg.add_header("Reply-to", reply_to)

            s.sendmail(from_address, to_address, msg.as_string())

        finally:
            if s:
                s.quit()


def send_template(template_name, to_address, from_address=default_from, subject=None,
                  reply_to=None, use_reply_to=True, **kwargs):

    text_template = _get_template(template_name, "txt")
    html_template = _get_template(template_name, "html")

    text_msg = None
    if text_template:
        text_msg = text_template.render(**kwargs)

    html_msg = None
    if html_template:
        html_msg = html_template.render(**kwargs)

    send_message(to_address, from_address=from_address, subject=subject,
                 reply_to=reply_to, use_reply_to=use_reply_to, text_msg=text_msg, html_msg=html_msg)



import json

from collections import Iterable

from dorthy.json import jsonify
from dorthy.enum import DeclarativeEnum
from dorthy.utils import create_frozenset, native_str


class AccessVotes(DeclarativeEnum):

    Abstain = 1,
    Denied = 2,
    Granted = 3,


class AccessDeniedError(Exception):
    pass


class AuthorityJSONSerializer(object):

    def deserialize(self, data):
        dct = data if isinstance(data, dict) else json.loads(native_str(data))
        authority = dct["authority"]
        permission = dct.get("permission", None)
        return Authority(authority, permission)

    def serialize(self, authority):
        return jsonify(authority)


class Authority(object):
    """Authority object that represents a permission within the system.
    An authority can additionally have a permission associated
    with it to further narrow it's definition.
    """

    def __init__(self, authority, permission=None):
        self.__authority = authority
        self.__permission = permission

    def __eq__(self, other):
        if not other:
            return False
        if type(other) is not Authority:
            return False
        if self is other:
            return True
        if self.authority == other.authority and \
                self.permission == other.permission:
            return True
        return False

    def __hash__(self):
        if not hasattr(self, "_hash"):
            if not self:
                value = 0
            else:
                value = hash(self.__authority)
                if self.__permission is not None:
                    value = (value << 1) ^ hash(self.__permission)
            self._hash = value
        return self._hash

    @property
    def authority(self):
        return self.__authority

    @property
    def permission(self):
        return self.__permission

    def with_permission(self, permission):
        if self.__permission:
            raise ValueError("Cannot redefine an authorities permission.")
        self.__permission = permission
        return self


def authority(auth, permission=None):
    """A helper method that provides readability to
    the authorized decorators.

    Args:
        auth: the authority which can be defined as a string or as an
        Authority object
        permission: the permission to set for the authority

    Returns:
        an AuthorityExpression with the authority set
    """
    if isinstance(auth, Authority):
        return AuthorityExpression(auth)
    elif isinstance(auth, str):
        return AuthorityExpression(Authority(auth, permission))
    else:
        raise ValueError("Invalid authority type. Only supports string and Authority.")


def member(group):
    """A helper method that provides readability to
    the authorized decorators.

    Args:
        group: the group name as a string

    Returns:
        a GroupExpression with the group set
    """
    return GroupExpression(group)


class Expression(object):

    def __and__(self, other):
        return AndExpression(self, other)

    def __or__(self, other):
        return OrExpression(self, other)

    def __neg__(self):
        return NotExpression(self)

    def apply(self, authentication, attribute, **options):
        raise NotImplementedError


class NotExpression(Expression):

    def __init__(self, expression):
        self.expression = expression

    def apply(self, authentication, attribute, **options):
        return not self.expression.apply(authentication, attribute, **options)


class LogicExpression(Expression):

    def __init__(self, lh_expression, rh_expression):
        self.lh_expression = lh_expression
        self.rh_expression = rh_expression


class AndExpression(LogicExpression):

    def apply(self, authentication, attribute, **options):
        return self.lh_expression.apply(authentication, attribute, **options) and \
            self.rh_expression.apply(authentication, attribute, **options)


class OrExpression(LogicExpression):

    def apply(self, authentication, attribute, **options):
        return self.lh_expression.apply(authentication, attribute, **options) or \
            self.rh_expression.apply(authentication, attribute, **options)


class ValueExpression(Expression):

    def __init__(self, value):
        self.value = value


class AuthorityExpression(ValueExpression):

    def apply(self, authentication, attribute, **options):
        return self.value in authentication.get_authorities()


class GroupExpression(ValueExpression):

    def apply(self, authentication, attribute, **options):
        group_names = {g.name for g in authentication.get_groups() if g.security_group}
        return True if self.value in group_names else False


class GroupVoter(object):
    """A security group voter that votes based upon
    group membership.
    """

    def __init__(self, groups):
        """Initializes a GroupAccessVoter

        Args:
            groups (str or Iterable): the group names to check access against
        """
        self.__groups = create_frozenset(groups)

    def supports(self, expression, attribute=None):
        """Determines if voter is supported based upon the
        attribute it is tied to.

        Args:
            expression: the security expression
            attribute (object): the attribute the expression is tied to

        Returns:
            always returns True
        """
        return True

    def vote(self, authentication, expression, attribute=None, **options):
        """Votes on access rights by group membership.

        Args:
            authentication (Authentication): the authentication
            expression: the security expression
            attribute (object): the calling attribute if avaiplatle
            options: key-work options

        Returns:
            AccessVotes.Granted if the authentication contains one of the given groups
        """
        if authentication.get_groups():
            group_names = {g.name for g in authentication.get_groups() if g.security_group}
            intersect = self.__groups & group_names
            return AccessVotes.Granted if intersect else AccessVotes.Denied
        else:
            return AccessVotes.Abstain


class ExpressionVoter(object):
    """A security expression voter that processes
    a security expression and votes on access based upon
    that expression and the authority.
    """

    def supports(self, expression, attribute=None):
        """Determines if voter is supported based upon the
        attribute it is tied to.

        Args:
            expression: the security expression
            attribute (object): the attribute the expression is tied to

        Returns:
            always returns True
        """
        return True

    def vote(self, authentication,  expression, attribute=None, **options):
        """Votes on access rights by checking a principal's granted authorities
         against the security expression.

        Args:
            authentication (Authentication): the authentication
            expression: the security expression
            attribute (object): the calling attribute if avaiplatle
            options: key-work options

        Returns:
            AccessVote. AccessVotes.Granted if the authentication has been granted the
            necessary authority otherwise AccessVotes.Denied.  If the expression is
            not a security expression check expression than AccessVotes.Abstain is returned.
            If the expression is None then returns AccessVotes.Granted.
        """
        if expression and isinstance(expression, Expression):
            if expression.apply(authentication, attribute, **options):
                return AccessVotes.Granted
            else:
                return AccessVotes.Denied
        return AccessVotes.Abstain


class BaseDecisionManager(object):

    def __init__(self, allow_all_abstain=False):
        self.__allow_all_abstain = allow_all_abstain

    @property
    def allow_all_abstain(self):
        return self.__allow_all_abstain

    def supports(self, expression, attribute=None):
        return True

    def decide(self, authentication, expression, attribute=None, **options):
        raise NotImplementedError()


class UnanimousDecisionManager(BaseDecisionManager):
    """A decision manager that requires all voters to either grant access or
    abstain -- one AccessVote.Denied will deny access for the authentication.
    """

    def __init__(self, voters, allow_all_abstain=False):
        """Creates a new decision manager.

        Args:
            voters: the voters used to make access decisions.  Accepts either a
            single voter or an iterable collection of voters.
        """
        assert voters, "Cannot initialize without voters."
        super().__init__(allow_all_abstain)
        if not isinstance(voters, Iterable):
            self.__voters = [voters]
        else:
            self.__voters = voters

    def decide(self, authentication, expression, attribute=None, **options):
        """Decides whether the authentication is allowed to access the attribute given
        the security expression.

        Args:
            authentication (Authentication): the authentication
            expression: the security expression
            attribute (object): the calling attribute if avaiplatle
            options: key-work options

        Raises:
            AccessDeniedError: authentication is denied access
        """
        if not self.supports(expression, attribute):
            raise AccessDeniedError()

        if authentication:
            abstains = 0
            for voter in self.__voters:
                if voter.supports(expression, attribute):
                    vote = voter.vote(authentication, expression, attribute=attribute, **options)
                    if vote == AccessVotes.Denied:
                        raise AccessDeniedError()
                    if vote == AccessVotes.Abstain:
                        abstains += 1
                else:
                    abstains += 1

            if not self.allow_all_abstain and abstains == len(self.__voters):
                raise AccessDeniedError()

        else:
            raise AccessDeniedError("Not authenticated")


class SuperUserDecisionManager(UnanimousDecisionManager):
    """A decision manager that grants access to a super user.
    """

    def __init__(self, super_voter, voters, allow_all_abstain=False):
        """Creates a new decision manager.

        Args:
            super_voter: a voter that used to make decisions about a super user
            voters: the voters used to make access decisions.  Accepts either a
            single voter or an iterable collection of voters.
        """
        assert super_voter, "Cannot initialize without a super voter."
        super().__init__(voters, allow_all_abstain)
        self.__super_voter = super_voter

    def decide(self, authentication, expression, attribute=None, **options):
        """Decides whether the authentication is allowed to access the attribute given
        the security expression.

        Args:
            authentication (Authentication): the authentication
            expression: the security expression
            attribute (object): the calling attribute if available
            options: key-work options

        Raises:
            AccessDeniedError: authentication is denied access
        """

        if authentication:
            # check to see if admin access is granted
            if self.__super_voter.supports(expression, attribute) and \
                    self.__super_voter.vote(
                        authentication, expression, attribute=attribute, **options) == AccessVotes.Granted:
                return

            # check with the rest of the voters
            super().decide(authentication, expression, attribute=attribute, **options)
        else:
            raise AccessDeniedError("Not authenticated")



"""Enum implementation originally from zzzeek:
http://techspot.zzzeek.org/2011/01/14/the-enum-recipe/

Our db enum type does not use sqlalchemy's built-in enum type
instead we just use a string.  We do not have consistency checks
within the database.
"""

from sqlalchemy.types import SchemaType, TypeDecorator, Integer, String


class EnumSymbol(object):
    """Define a fixed symbol tied to a parent class."""

    def __init__(self, cls_, name, value, description=None):
        self.cls_ = cls_
        self.__name = name
        self.__value = value
        self.__description = description

    def __eq__(self, other):
        return id(self) == id(other)

    def __hash__(self):
        return id(self)

    def __reduce__(self):
        """Allow unpickling to return the symbol
        linked to the DeclarativeEnum class."""
        return getattr, (self.cls_, self.name)

    def __iter__(self):
        return iter([self.value, self.description])

    def __repr__(self):
        if self.__description:
            return "<{}, {}, {}>".format(self.name, self.value, self.description)
        else:
            return "<{}, {}>".format(self.name, self.value)

    def __str__(self):
        return str(self.__value)

    def _as_dict(self):
        return {
            "name": self.name,
            "value": self.value,
            "description": self.description
        }

    @property
    def name(self):
        return self.__name

    @property
    def value(self):
        return self.__value

    @property
    def description(self):
        return self.__description


class EnumMeta(type):
    """Generate new DeclarativeEnum class.

    Only supports int and string type as values of the enum
    """

    def __init__(cls, classname, bases, dict_):

        cls._registered = registered = cls._registered.copy()

        for k, v in dict_.items():

            if k[0] == "_":
                continue

            value = description = None
            if isinstance(v, tuple) and len(v) > 0:
                assert isinstance(v[0], str) or isinstance(v[0], int), \
                    "Declarative Enum only supports string and ints as values."
                value = v[0]
                # if defined as a tuple w/o description use the key
                description = v[1] if len(v) > 1 else k
            elif isinstance(v, str) or isinstance(v, int):
                value = v

            if value is not None:
                if cls.int_enum() and not isinstance(value, int):
                    raise ValueError("Enum can only contain integer values.")

                # normalize all keys to strings
                value_str = str(value)
                if value_str in registered:
                    raise ValueError("Declarative Enum only supports unique values.")

                sym = registered[value_str] = EnumSymbol(cls, k, value, description=description)
                setattr(cls, k, sym)

        type.__init__(cls, classname, bases, dict_)


class DeclarativeEnum(metaclass=EnumMeta):
    """Declarative enumeration that will except
    String and int values"""

    _registered = dict()

    @classmethod
    def convert(cls, value):
        """
        Converts the given value to the enum value

        Returns:
            the enum for the given value
        Raises:
            ValueError: the enum value does not exist in the enum
        """
        # convert non-string values
        value = str(value)
        try:
            return cls._registered[value]
        except KeyError:
            raise ValueError("Invalid value for {}: {}".format(cls.__name__, value))

    @classmethod
    def values(cls):
        """
        Returns all the EnumSymbol values for the given enum.

        Returns:
            all EnumSymbol values for the given enum
        """
        return cls._registered.values()

    @classmethod
    def keys(cls):
        """
        Returns all the keys for the given enum.

        Returns:
            all EnumSymbol keys for the given enum
        """
        return cls._registered.keys()

    @classmethod
    def int_enum(cls):
        return False

    @classmethod
    def db_type(cls):
        """
        Generates the SQLAlchemy DB type for the given enum

        Returns:
            a DeclarativeEnumType for the enum
        """
        return DeclarativeEnumType(cls)


class IntDeclarativeEnum(DeclarativeEnum):
    """A declarative enumeration with that will
    only except int values.
    """

    # needed to differentiate the class because the definition of the class
    # is not available to the meta class when DeclarativeEnum is first called
    @classmethod
    def int_enum(cls):
        return True


class DeclarativeEnumType(SchemaType, TypeDecorator):
    """SQLAlchemy DeclarativeEnum type

    This implementation either stores all enum values as strings in the DB
    or can be configured to store values as integers.
    """

    def __init__(self, enum):
        super().__init__()
        if enum.int_enum():
            self.impl = Integer()
        else:
            for v in enum.values():
                v_str = str(v.value)
                if len(v_str) > 10:
                    raise ValueError("Enum keys cannot be greater than 10 characters. Enum: {} Key: {}".
                                     format(enum.__name__, v_str))
            self.impl = String(length=10)

        self.enum = enum

    def copy(self, **kwargs):
        return DeclarativeEnumType(self.enum)

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        if self.enum.int_enum():
            return value.value
        else:
            return str(value.value)

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        if not self.enum.int_enum():
            value = value.strip()
        return self.enum.convert(value)

    def __repr__(self):
        """Defined to override default behavior for alembic generation
        see: https://bitbucket.org/zzzeek/alembic/issue/78/autogenerate-does-not-detect-decorated
        """
        if self.enum.int_enum():
            return "Integer"
        else:
            return "String(length=10)"



import base64
import logging
import pickle
import sys

from collections import MutableMapping
from time import time
from uuid import uuid4

from tornado.escape import json_decode

from dorthy.json import jsonify

logger = logging.getLogger(__name__)

DEFAULT_SESSION_TIMEOUT = 86400 * 1


class Session(MutableMapping):
    """Provides a session data structure that provides dictionary
    access.  Provides JSON serialization and deserialization methods.
    """

    _SYS_ENCODING = sys.getdefaultencoding()

    def __init__(self, session_id, timeout=DEFAULT_SESSION_TIMEOUT, update_access=True):
        self.__session_id = session_id
        self.__created = time()
        self.__last_accessed = self.__created
        self.__new = True
        self.__modified = False
        self.__valid = True
        self.__data = dict()
        self.timeout = timeout
        self.update_access = update_access

    def __contains__(self, key):
        return key in self.__data

    def __len__(self):
        return len(self.__data)

    def __getitem__(self, key):
        return self.__data[key]

    def __setitem__(self, key, value):
        self.__data[key] = value
        self.__modified = True

    def __delitem__(self, key):
        del self.__data[key]
        self.__modified = True

    def __iter__(self):
        return iter(self.__data)

    def clear(self):
        self.__data.clear()
        self.__modified = True

    def pop(self, key, *args):
        self.__modified = self.__modified or key in self.__data
        return self.__data.pop(key, *args)

    @property
    def created(self):
        return self.__created

    @property
    def session_id(self):
        return self.__session_id

    @property
    def data(self):
        return dict(self.__data)

    def expired(self):
        if self.timeout <= 0:
            expired = False
        elif self.last_accessed == 0:
            expired = self.created + self.timeout < time()
        else:
            expired = self.last_accessed + self.timeout < time()
        return expired

    @property
    def last_accessed(self):
        return self.__last_accessed

    def _update_accessed(self):
        if self.update_access:
            self.__last_accessed = time()

    @property
    def modified(self):
        return self.__modified

    @property
    def is_new(self):
        return self.__new

    @property
    def valid(self):
        return self.__valid

    def encode(self):
        d = self._as_dict()
        # pickle data elements
        if d["data"]:
            pickled = pickle.dumps(self.__data)
            d["data"] = base64.standard_b64encode(pickled).decode(self._SYS_ENCODING)
        return jsonify(d)

    def _as_dict(self):
        d = dict()
        d["session_id"] = self.__session_id
        d["created"] = self.__created
        d["last_accessed"] = self.__last_accessed
        d["timeout"] = self.timeout
        d["update_access"] = self.update_access
        d["data"] = self.__data
        return d

    @classmethod
    def decode(cls, data):
        d = json_decode(data)
        self = cls.__new__(cls)
        self.__session_id = d["session_id"]
        self.__created = d["created"]
        self.__last_accessed = d["last_accessed"]
        self.__new = False
        self.__modified = False
        self.timeout = d["timeout"]
        if not self.timeout:
            self.timeout = DEFAULT_SESSION_TIMEOUT
        self.update_access = d["update_access"]
        self.__valid = True
        if "data" not in d:
            self.__data = dict()
        elif d["data"]:
            data = d["data"]
            # decode pickled data
            b = data.encode(self._SYS_ENCODING)
            pickled = base64.standard_b64decode(b)
            self.__data = pickle.loads(pickled)
        return self

    def invalidate(self):
        self.__valid = False
        self.__modified = True


class BaseSessionStore(object):

    @staticmethod
    def generate_session_id():
        return uuid4().hex

    def load(self, session_id):
        pass

    def save(self, session):
        if not session.valid:
            self._delete(session.session_id)
        else:
            self._store_session(session)

    def _delete(self, session_id):
        pass

    def _store_session(self, session):
        pass

    def _validate_session(self, session):
        if session and session.expired():
            self._delete(session.session_id)
            session = None
        return session


class InMemorySessionStore(BaseSessionStore):

    def __init__(self):
        self.__store = dict()

    def load(self, session_id):
        return self._validate_session(self.__store.get(session_id, None))

    def _delete(self, session_id):
        del self.__store[session_id]

    def _store_session(self, session):
        if session.valid:
            session._update_accessed()
            self.__store[session.session_id] = session
        else:
            self._delete(session.session_id)



import redis
import time

from dorthy.settings import config
from dorthy.utils import native_str


client = redis.StrictRedis(host=config.redis.server,
                           port=config.redis.port,
                           db=config.redis.db)


def create_key(prefix, key, decode_key=False):
    key = key if not decode_key else native_str(key)
    return "{}:{}".format(prefix, key)


def get_cached_value(key, ttl):
    """
    Gets a cached value and determines whether or not it has expired
    according to its time to live (ttl)

    Args:
        key: key under which data is stored
        ttl: time to live for key (in seconds)
    Returns:
        (cached_value, expired)
    """
    cached_value = client.hget(key, 'data')
    timestamp = client.hget(key, 'ts')

    if timestamp and time.time() - float(timestamp) > ttl:
        expired = True
    else:
        expired = False
    return cached_value, expired


def cache_value(key, value):
    """
    Caches a value along with current timestamp

    Args:
        key: key under which data is stored
        value: str or number value ready to be stored
    Returns:
        Redis return value
    """
    assert isinstance(value, (str, int, float, complex))
    return client.hmset(key, {'data': value, 'ts': time.time()})